import * as React from 'react';
import { EncTypeSerializerMap, serializeBody, SerializerOptions } from '../../utils/serialization';
import { ENCTYPE_MULTIPART_FORM_DATA } from '../../common/enctypes';
import { AllowedClientMethod, FormDerivedElement } from '../common';
import { METHODS_WITH_BODY } from '../../common/constants';

export interface UseFormFetchParams {
	clientAction?: string;
	clientHeaders?: HeadersInit;
	clientMethod: AllowedClientMethod;
	encType: string;
	invalidate?: (...args: unknown[]) => unknown;
	refresh?: (response: Response) => void;
	encTypeSerializers?: EncTypeSerializerMap;
	responseEncType: string;
	serializerOptions?: SerializerOptions;
	onSubmit?: React.FormEventHandler<FormDerivedElement>;
}

export const useFormFetch = ({
	clientAction,
	invalidate,
	clientHeaders,
	responseEncType,
	encType,
	clientMethod,
	encTypeSerializers,
	serializerOptions,
	refresh,
	onSubmit,
}: UseFormFetchParams) => {
	const handleSubmit: React.FormEventHandler<
		FormDerivedElement
	> = React.useCallback(async (event) => {
		event.preventDefault();
		const { submitter } = event.nativeEvent as unknown as { submitter?: HTMLElement };

		if (clientAction) {
			invalidate?.();
			const headers: HeadersInit = {
				...(clientHeaders ?? {}),
				Accept: responseEncType,
			};
			if (encType !== ENCTYPE_MULTIPART_FORM_DATA) {
				// browser automatically generates content-type header for multipart/form-data
				(headers as unknown as Record<string, string>)['Content-Type'] = encType;
			}

			const fetchInit: RequestInit = {
				method: clientMethod.toUpperCase(),
				headers,
			};

			if (
				METHODS_WITH_BODY.includes(clientMethod.toUpperCase() as typeof METHODS_WITH_BODY[number])
			) {
				fetchInit.body = serializeBody({
					form: event.currentTarget,
					encType,
					serializers: encTypeSerializers,
					serializerOptions: {
						...serializerOptions,
						submitter,
					},
				});
			}
			const response = await fetch(clientAction, fetchInit);
			refresh?.(response);
		}

		onSubmit?.(event);
	}, [
		clientAction,
		invalidate,
		clientHeaders,
		responseEncType,
		encType,
		clientMethod,
		encTypeSerializers,
		serializerOptions,
		refresh,
		onSubmit,
	]);

	return React.useMemo(() => ({
		handleSubmit,
	}), [
		handleSubmit,
	]);
};
