import { ParsedUrlQuery } from 'querystring';
import { GetServerSideProps, GetServerSidePropsContext, PreviewData } from 'next';
import { deserialize } from 'seroval';
import { NextApiRequest, NextApiResponse } from '../common/types';
import { DEFAULT_METHOD } from './constants';
import { getBody } from '../utils/request';
import { ENCTYPE_APPLICATION_JSON, ENCTYPE_APPLICATION_OCTET_STREAM } from '../common/enctypes';
import { METHODS_WITH_BODY, REQUEST_ID_FORM_KEY, DEFAULT_ENCODING } from '../common/constants';
import { retrieveCache } from './cache';

export type DestinationGetServerSideProps<
	Props extends Record<string, unknown> = Record<string, unknown>,
	Params extends ParsedUrlQuery = ParsedUrlQuery,
	Preview extends PreviewData = PreviewData,
> = (
	actionReq: NextApiRequest,
	actionRes: NextApiResponse,
	context: GetServerSidePropsContext<Params, Preview>,
) => ReturnType<GetServerSideProps<Props, Params, Preview>>;

export interface DestinationWrapperOptions {
	fn?: DestinationGetServerSideProps;
	requestIdKey?: string;
}

export const getServerSideProps = (
	options = {} as DestinationWrapperOptions,
): GetServerSideProps => async (ctx) => {
	const req: NextApiRequest = {
		query: {
			...ctx.query,
			...(ctx.params ?? {}),
		},
		body: null,
	};
	const { method = DEFAULT_METHOD } = ctx.req;

	if (METHODS_WITH_BODY.includes(method.toUpperCase() as typeof METHODS_WITH_BODY[number])) {
		const body = await getBody(ctx.req);
		req.body = body.toString(DEFAULT_ENCODING);
	}

	const res: NextApiResponse = {};
	const resRequestId = ctx.query[
		options.requestIdKey ?? REQUEST_ID_FORM_KEY
	] as string | undefined;
	if (resRequestId) {
		const {
			statusCode,
			statusMessage,
			contentType,
			data: resBody,
		} = await retrieveCache(resRequestId);
		ctx.res.statusCode = statusCode;
		if (typeof statusMessage !== 'undefined') {
			ctx.res.statusMessage = statusMessage;
		}
		if (resBody) {
			if (contentType === ENCTYPE_APPLICATION_JSON) {
				res.body = deserialize(resBody as string);
			} else if (contentType === ENCTYPE_APPLICATION_OCTET_STREAM) {
				res.body = deserialize(resBody as string);
			} else {
				const c = console;
				c.warn('Could not parse response body, returning nothing');
				res.body = null;
			}
		} else {
			res.body = null;
		}
	}

	const gspResult = (
		typeof options?.fn === 'function'
			? await options.fn(req, res, ctx)
			: {
				props: {},
			}
	);

	if ('props' in gspResult) {
		return {
			...gspResult,
			props: {
				...gspResult.props,
				req,
				res,
			},
		};
	}

	// redirect/not found will be treated as default behavior
	return gspResult;
};
